"""
ROAD2Timeline Plugin
Contributed by Ryan Marcotte Cobb (Secureworks)

Copyright 2022 - MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import argparse
import sqlalchemy

import roadtools.roadlib.metadef.database as database

from pathlib import Path
from typing import Dict, Optional

# Currently, there is no good way to specify
# plugin-specific imports. Therefore, we need
# to manually check that the necessary modules
# are present in the environment.

try:
    import yaml
    import pandas as pd
    import numpy as np
    HAS_RTT_MODULES = True
except (ModuleNotFoundError, ImportError) as exc:
    HAS_RTT_MODULES = False


DESCRIPTION = "Timeline analysis of Azure AD objects"


def create_args_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
    )
    parser.add_argument(
        "-d",
        "--database",
        action="store",
        help="Database file. Can be the local database name for SQLite, or an SQLAlchemy "
        "compatible URL such as postgresql+psycopg2://dirkjan@/roadtools",
        default="roadrecon.db",
    )
    return parser


def add_args(parser: argparse.ArgumentParser):
    """Plugin-specific arguments"""
    parser.add_argument(
        "-t",
        "--template-file",
        action="store",
        help="File containing string templates to translate Azure AD objects into a timeline \
            entry. Defaults to `road2timeline.yaml` in the current working directory.",
        default="road2timeline.yaml",
    )
    parser.add_argument(
        "-f",
        "--output-file",
        action="store",
        help="File to save timeline outputs. Output format determined by file extension. Supported extensions include: [csv, pickle, jsonl, parquet]",
        default="timeline-output.jsonl",
    )

if HAS_RTT_MODULES:
    def populate_timeline_entry(
        row: pd.Series, timeline_entry_templates: Dict[str, Dict[str, str]]
    ) -> str:
        """Attempts to populate a timeline entry template with
        the relevant fields from a row in the database.

        Parameters
        ----------
        row : pd.Series
            A row from a DataFrame, passed by `.apply(..., axis=1)`
        timeline_entry_templates : Dict[str, Dict[str, str]]
            Dictionary of template strings organized by table and
            timestamp column name.

        Returns
        -------
        str
            Templated timeline event message
        """

        table_templates = timeline_entry_templates.get(row._table_name, {})
        template_text = table_templates.get(row._timestamp_column)
        if template_text:
            try:
                return template_text.format(**row.to_dict())
            except Exception as exc:
                print(f"There was a problem parsing the message: {str(exc)}")
                return f"Error parsing template {row._table_name}.{row._timestamp_column}: Object ID - {row._object_id}"
        else:
            return f"No template found for {row._table_name}.{row._timestamp_column}: Object ID - {row._object_id}"


    def to_dataframe(
        session: sqlalchemy.orm.session.Session,
        table: sqlalchemy.orm.DeclarativeMeta,
    ) -> pd.DataFrame:
        """Reads sqlite table and converts to pd.DataFrame.
        This also converts the string values in datetime columns
        into their equivalent Python `datetime` representation.

        Parameters
        ----------
        session : sqlalchemy.orm.session.Session
            SQLAlchemy session with the ROADtools database
        table : sqlalchemy.ext.declarative.api.DeclarativeMeta
            An object representing the SQL table in the ROADtools
            database

        Returns
        -------
        pd.DataFrame
            All data from the table converted to Python data
            types
        """
        rows = session.query(table).all()
        if not rows:
            return pd.DataFrame()

        df = pd.json_normalize(
            [
                {
                    **row._asdict(),
                    "_object_id": getattr(row, "objectId", None)
                    or getattr(row, "id", None)
                    or "Identifier Not Found",
                }
                for row in rows
            ]
        ).assign(_table_name=table.name)

        for column in table.columns:
            # Attempts to coerce sqlite datetime columns
            # into a python Datetime object.
            if type(column.type) == sqlalchemy.sql.sqltypes.DATETIME:
                if column.name in df.columns:
                    df[column.name] = df[column.name].apply(pd.to_datetime, errors="coerce")

        return df


    def copy_dataframe_by_col(df: pd.DataFrame, col: str) -> pd.DataFrame:
        """Returns a copy of a `pd.DataFrame` with additional columns
        for a given datetime field.

        Parameters
        ----------
        df : pd.DataFrame
            Source DataFrame
        col : str
            Timestamp column of interest

        Returns
        -------
        pd.DataFrame
            Copy of the source DataFrame, but with additional
            columns detailing the column of interest and dropping
            all rows where that timestamp column is not null
        """
        df = df.copy()
        df["_timestamp"] = df[col]
        df["_timestamp_column"] = col
        return df[df[col].notnull()]


def main(args: Optional[argparse.Namespace] = None) -> Path:

    if args is None:
        parser = create_args_parser()
        args = parser.parse_args()

    # Connect to the database
    db_url = database.parse_db_argument(args.database)
    engine = database.init(dburl=db_url)
    session = database.get_session(engine)

    # Do some reflection to grab the tables
    # and their respective column types
    db_metadata = sqlalchemy.MetaData(bind=engine, schema="main")
    db_metadata.reflect()

    if not HAS_RTT_MODULES:
        print('Error importing required modules for road2timeline. Make sure pyyaml, numpy and pandas are installed.')
        return

    # Find the templates file which is used
    # to populate timeline entries from the
    # source rows.
    templates_fp = Path(args.template_file)
    if not templates_fp.exists():
        print(
            f"Timeline entry template file {templates_fp} not found, defaulting to built-in templates"
        )
        templates_fp = Path(__file__).with_suffix(".yaml")

    print(f"Loading timeline entry templates from {templates_fp}")
    timeline_entry_templates = yaml.safe_load(
        templates_fp.read_text(encoding="utf-8", errors="replace")
    )

    # For each table, convert its rows into
    # a pandas DataFrame, then create n copies
    # of the DataFrame for each datetime column
    # in the DataFrame with additional metadata.
    dataframes = []
    for table in db_metadata.sorted_tables:
        df = to_dataframe(session, table)
        for col in df.select_dtypes(include=[np.datetime64]).columns:
            dataframes.append(copy_dataframe_by_col(df, col))

    # Transform the rows of the DataFrame into
    # a human-readable string that will be used
    # as the entry in the timeline `_message` field.
    # Then sort the timeline.
    timeline = (
        pd.concat(dataframes, ignore_index=True)
        .assign(
            _message=lambda x: x.apply(
                populate_timeline_entry, args=(timeline_entry_templates,), axis=1
            )
        )
        # TODO: Decide if a better column naming scheme is required.
        # See the actual [`log2timeline` docs](https://plaso.readthedocs.io/en/latest/sources/user/Output-and-formatting.html)
        # for column names that would be compatible with their data model.
        #
        # Hacky way get these columns sorted first
        .set_index(["_timestamp_column", "_table_name", "_message", "_object_id"])
        # The remaining columns are then sorted alphabetically
        .sort_index(axis=1)
        .reset_index()
        # Set the index datetime
        .set_index("_timestamp")
        # Sort by timestamp
        .sort_index()
        # .reset_index()
    )

    # Infer the output format based on the
    # file extension of the `--output-file`
    # argument.
    output_file = Path(args.output_file)

    if output_file.suffix == ".jsonl":
        timeline.to_json(output_file, orient="records", lines=True, default_handler=str)
    elif output_file.suffix == ".pickle":
        timeline.to_pickle(str(output_file))
    elif output_file.suffix == ".csv":
        timeline.to_csv(output_file)
    elif output_file.suffix == ".parquet":
        timeline.to_parquet(output_file)
    else:
        raise ValueError(
            f"Unable to determine output format \
            for `--output-file` argument {args.output_file}. \
            Filename must end with one of: [jsonl, pickle, csv, parquet]"
        )

    print(f"Timeline saved to file {output_file.resolve()}")

    return output_file


if __name__ == "__main__":
    main()



import argparse
import sys
import os
import importlib
from roadtools.roadlib.auth import Authentication
from roadtools.roadrecon.gather import getargs as getgatherargs
RR_HELP = '''ROADrecon - The Azure AD exploration tool.
By @_dirkjan - dirkjanm.io

To get started, use one of the subcommands. Each command has a help feature (roadrecon <command> -h).

1. Authenticate to Azure AD
roadrecon auth <options>

2. Gather all information
roadrecon gather <options>

3. Explore the data or export it to a specific format using a plugin
roadrecon gui
roadrecon plugin -h
'''

def check_database_exists(path):
    '''
    Small sanity check to see if the specified database exists.
    Otherwise SQLAlchemy creates it without data and throws errors later, which does
    not help anyone
    '''
    found = False
    if ':/' in path:
        found = True
    else:
        if path[0] != '/':
            found = os.path.exists(os.path.join(os.getcwd(), path))
        else:
            found = os.path.exists(path)
    if not found:
        raise Exception('The database file {0} was not found. Please make sure it exists'.format(path))

def main():
    # Primary argument parser
    parser = argparse.ArgumentParser(add_help=True, description=RR_HELP, formatter_class=argparse.RawDescriptionHelpFormatter)
    # Add subparsers for modules
    subparsers = parser.add_subparsers(dest='command')

    # Construct authentication module options
    auth = Authentication()
    auth_parser = subparsers.add_parser('auth', help='Authenticate to Azure AD')
    auth.get_sub_argparse(auth_parser, for_rr=True)

    # Construct gather module options (imported from gather module)
    gather_parser = subparsers.add_parser('gather', aliases=['dump'], help='Gather Azure AD information')
    getgatherargs(gather_parser)

    # Construct GUI options
    gui_parser = subparsers.add_parser('gui', help='Launch the web-based GUI')
    gui_parser.add_argument('-d',
                            '--database',
                            action='store',
                            help='Database file. Can be the local database name for SQLite, or an SQLAlchemy compatible URL such as postgresql+psycopg2://dirkjan@/roadtools',
                            default='roadrecon.db')
    gui_parser.add_argument('--debug',
                            action='store_true',
                            help='Enable flask debug')
    gui_parser.add_argument('--profile',
                            action='store_true',
                            help='Enable flask profiler')

    # Construct plugins module options
    plugin_parser = subparsers.add_parser('plugin', help='Run a ROADrecon plugin')
    plugins = plugin_parser.add_subparsers(dest='plugin')

    # If you added a new (custom) plugin to the /plugins/ directory, add it to the list here
    # with a short description
    plugins_list = {
        'policies': 'Parse conditional access policies',
        'bloodhound': 'Export Azure AD data to a custom BloodHound version',
        'xlsexport': 'Export data to an Excel file',
        'road2timeline': 'Generate a forensic timeline from Azure AD object timestamps',
        # 'grep': 'Export grep-compatible lists'
    }

    # Iterate over plugins
    for plugin, description in plugins_list.items():
        # Import the plugin
        plugin_module = importlib.import_module('roadtools.roadrecon.plugins.{}'.format(plugin))
        pparser = plugins.add_parser(plugin, description=plugin_module.DESCRIPTION, help=description)

        # Every plugin uses at least the database, so add that option
        pparser.add_argument('-d',
                             '--database',
                             action='store',
                             help='Database file. Can be the local database name for SQLite, or an SQLAlchemy compatible URL such as postgresql+psycopg2://dirkjan@/roadtools',
                             default='roadrecon.db')
        plugin_module.add_args(pparser)


    args = parser.parse_args()

    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(1)
        return

    args = parser.parse_args()
    if args.command == 'auth':
        auth.parse_args(args)
        res = auth.get_tokens(args)
        # Could probably be shortened but older versions of roadlib may
        # return None and I just want to make sure that doesn't break here
        if res is False:
            return
        auth.save_tokens(args)
    elif args.command == 'gui':
        from roadtools.roadrecon.server import main as servermain
        check_database_exists(args.database)
        servermain(args)
    elif args.command == 'gather' or args.command == 'dump':
        from roadtools.roadrecon.gather import main as gathermain
        gathermain(args)
    elif args.command == 'plugin':
        # Dynamic import
        plugin_module = importlib.import_module('roadtools.roadrecon.plugins.{}'.format(args.plugin))
        check_database_exists(args.database)
        plugin_module.main(args)
if __name__ == '__main__':
    main()



from setuptools import setup
setup(name='roadrecon',
      version='1.1.4',
      description='Azure AD recon for red and blue',
      author='Dirk-jan Mollema',
      author_email='dirkjan@dirkjanm.io',
      url='https://github.com/dirkjanm/ROADtools/',
      license='MIT',
      classifiers=[
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
      ],
      packages=['roadtools.roadrecon', 'roadtools.roadrecon.plugins'],
      package_data={'roadtools.roadrecon.plugins': ['*.yaml']},
      install_requires=[
          'roadlib>=0.17',
          'flask<3',
          'sqlalchemy<2',
          'marshmallow',
          'flask-sqlalchemy>=2.5',
          'flask-marshmallow',
          'flask-cors',
          'marshmallow-sqlalchemy>=0.22',
          'aiohttp',
          'openpyxl'
      ],
      zip_safe=False,
      include_package_data=True,
      entry_points={
          'console_scripts': ['roadrecon-gui=roadtools.roadrecon.server:main',
                              'roadrecon=roadtools.roadrecon.main:main']
      }
      )



'''
Export ROADrecon data into BloodHound's neo4j database
Uses code from aclpwn under an MIT license

Copyright (c) 2020 Dirk-jan Mollema

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''
import platform
import os
import json
import argparse
import sys
from roadtools.roadlib.metadef.database import ServicePrincipal, User, Group, DirectoryRole
import roadtools.roadlib.metadef.database as database
try:
    from neo4j import GraphDatabase
    from neo4j.exceptions import ClientError
    HAS_NEO_MODULE = True
except ModuleNotFoundError:
    try:
        from neo4j.v1 import GraphDatabase
        from neo4j.exceptions import ClientError
        HAS_NEO_MODULE = True
    except ModuleNotFoundError:
        HAS_NEO_MODULE = False

DESCRIPTION = '''
Export ROADrecon data into BloodHound's neo4j database.
Requires a custom version of the BloodHound interface to use, available at
https://github.com/dirkjanm/BloodHound-AzureAD
'''

BASE_LINK_QUERY = 'UNWIND $props AS prop MERGE (n:{0} {{objectid: prop.source}}) MERGE (m:{1} {{objectid: prop.target}}) MERGE (n)-[r:{2}]->(m)';

def add_edge(tx, aid, atype, bid, btype, linktype):
    q = BASE_LINK_QUERY.format(atype, btype, linktype)
    props = {'source':aid, 'target':bid}
    tx.run(q, props=props)

class BloodHoundPlugin():
    """
    Export data to BloodHounds neo4j database plugin
    """
    def __init__(self, session, dbhost, dbuser, dbpass):
        # SQLAlchemy session
        self.session = session
        # Neo4j driver
        self.driver = self.init_driver(dbhost, dbuser, dbpass)

    @staticmethod
    def init_driver(database, user, password):
        """
        Initialize neo4j driver
        """
        uri = "bolt://%s:7687" % database
        driver = GraphDatabase.driver(uri, auth=(user, password), encrypted=False)
        return driver

    @staticmethod
    def detect_db_config():
        """
        Detect bloodhound config, which is stored in appData.
        OS dependent according to https://electronjs.org/docs/api/app#appgetpathname
        """
        system = platform.system()
        if system == 'Windows':
            try:
                directory = os.environ['APPDATA']
            except KeyError:
                return (None, None)
            config = os.path.join(directory, 'BloodHound', 'config.json')
            try:
                with open(config, 'r') as configfile:
                    configdata = json.load(configfile)
            except IOError:
                return (None, None)

        if system == 'Linux':
            try:
                directory = os.environ['XDG_CONFIG_HOME']
            except KeyError:
                try:
                    directory = os.path.join(os.environ['HOME'], '.config')
                except KeyError:
                    return (None, None)
            config = os.path.join(directory, 'bloodhound', 'config.json')
            try:
                with open(config, 'r') as configfile:
                    configdata = json.load(configfile)
            except IOError:
                return (None, None)

        if system == 'Darwin':
            try:
                directory = os.path.join(os.environ['HOME'], 'Library', 'Application Support')
            except KeyError:
                return (None, None)
            config = os.path.join(directory, 'bloodhound', 'config.json')
            try:
                with open(config, 'r') as configfile:
                    configdata = json.load(configfile)
            except IOError:
                return (None, None)

        # If we are still here, we apparently found the config :)
        try:
            username = configdata['databaseInfo']['user']
        except KeyError:
            username = 'neo4j'
        try:
            password = configdata['databaseInfo']['password']
        except KeyError:
            password = None
        return username, password

    def main(self):
        """
        Main plugin logic. Simply connects to both databases and transforms the data
        """
        print('Connecting to neo4j')
        with self.driver.session() as neosession:
            print('Running queries')

            try:
                neosession.run('CREATE CONSTRAINT ON (c:AzureUser) ASSERT c.objectid IS UNIQUE')
                neosession.run('CREATE CONSTRAINT ON (c:AzureGroup) ASSERT c.objectid IS UNIQUE')
                neosession.run('CREATE CONSTRAINT ON (c:AzureRole) ASSERT c.objectid IS UNIQUE')
                neosession.run('CREATE CONSTRAINT ON (c:ServicePrincipal) ASSERT c.objectid IS UNIQUE')
            except ClientError as e:
                pass  # on neo4j 4, an error is raised when the constraint exists already

            for user in self.session.query(User):
                property_query = 'UNWIND $props AS prop MERGE (n:AzureUser {objectid: prop.sourceid}) SET n += prop.map'
                uprops = {
                    'name': user.userPrincipalName,
                    'displayname': user.displayName,
                    'enabled': user.accountEnabled,
                    'distinguishedname': user.onPremisesDistinguishedName,
                    'email': user.mail,
                }
                props = {'map': uprops, 'sourceid': user.objectId}
                if user.onPremisesSecurityIdentifier:
                    # uprops['onPremisesSecurityIdentifier'] = user.onPremisesSecurityIdentifier
                    props['onpremid'] = user.onPremisesSecurityIdentifier
                    property_query = 'UNWIND $props AS prop MERGE (n:AzureUser {objectid: prop.sourceid}) MERGE (m:User {objectid:prop.onpremid}) MERGE (m)-[r:SyncsTo {isacl:false}]->(n) SET n += prop.map'

                res = neosession.run(property_query, props=props)

            for sprinc in self.session.query(ServicePrincipal):
                property_query = 'UNWIND $props AS prop MERGE (n:ServicePrincipal {objectid: prop.sourceid}) SET n += prop.map'
                uprops = {
                    'name': sprinc.displayName,
                    'appid': sprinc.appId,
                    'publisher': sprinc.publisherName,
                    'displayname': sprinc.displayName,
                    'enabled': sprinc.accountEnabled,
                }
                props = {'map': uprops, 'sourceid': sprinc.objectId}
                res = neosession.run(property_query, props=props)
                for owneruser in sprinc.ownerUsers:
                    add_edge(neosession, owneruser.objectId, 'AzureUser', sprinc.objectId, 'ServicePrincipal', 'Owns')
                for ownersp in sprinc.ownerServicePrincipals:
                    add_edge(neosession, ownersp.objectId, 'ServicePrincipal', sprinc.objectId, 'ServicePrincipal', 'Owns')


            for group in self.session.query(Group):
                property_query = 'UNWIND $props AS prop MERGE (n:AzureGroup {objectid: prop.sourceid}) SET n += prop.map'
                uprops = {
                    'name': group.displayName,
                    'displayname': group.displayName,
                    'email': group.mail,
                }
                props = {'map': uprops, 'sourceid': group.objectId}
                if group.onPremisesSecurityIdentifier:
                    # uprops['onPremisesSecurityIdentifier'] = group.onPremisesSecurityIdentifier
                    props['onpremid'] = group.onPremisesSecurityIdentifier
                    property_query = 'UNWIND $props AS prop MERGE (n:AzureGroup {objectid: prop.sourceid}) MERGE (m:Group {objectid:prop.onpremid}) MERGE (m)-[r:SyncsTo {isacl:false}]->(n) SET n += prop.map'

                res = neosession.run(property_query, props=props)
                for memberuser in group.memberUsers:
                    add_edge(neosession, memberuser.objectId, 'AzureUser', group.objectId, 'AzureGroup', 'MemberOf')
                for membergroup in group.memberGroups:
                    add_edge(neosession, membergroup.objectId, 'AzureGroup', group.objectId, 'AzureGroup', 'MemberOf')
                for membersp in group.memberServicePrincipals:
                    add_edge(neosession, membersp.objectId, 'AzureGroup', group.objectId, 'ServicePrincipal', 'MemberOf')


            for role in self.session.query(DirectoryRole):
                property_query = 'UNWIND $props AS prop MERGE (n:AzureRole {objectid: prop.sourceid}) SET n += prop.map'
                uprops = {
                    'name': role.displayName,
                    'displayname': role.displayName,
                    'description': role.description,
                    'templateid': role.roleTemplateId
                }
                props = {'map': uprops, 'sourceid': role.objectId}
                res = neosession.run(property_query, props=props)
                for memberuser in role.memberUsers:
                    add_edge(neosession, memberuser.objectId, 'AzureUser', role.objectId, 'AzureRole', 'MemberOf')

                for memberuser in role.memberServicePrincipals:
                    add_edge(neosession, memberuser.objectId, 'ServicePrincipal', role.objectId, 'AzureRole', 'MemberOf')

        print('Done!')
        self.driver.close()

def add_args(parser):
    #DB parameters
    databasegroup = parser.add_argument_group("Database options (if unspecified they will be taken from your BloodHound config)")
    databasegroup.add_argument("--neodatabase", type=str, metavar="DATABASE HOST", default="localhost", help="The host neo4j is running on. Default: localhost")
    databasegroup.add_argument("-du", "--database-user", type=str, metavar="USERNAME", default="neo4j", help="Neo4j username to use")
    databasegroup.add_argument("-dp", "--database-password", type=str, metavar="PASSWORD", help="Neo4j password to use")


def main(args=None):
    if not HAS_NEO_MODULE:
        print('neo4j python module not found! Please install the module neo4j-driver first (pip install neo4j-driver)')
        sys.exit(1)
        return
    if args is None:
        parser = argparse.ArgumentParser(add_help=True, description='ROADrecon policies to HTML plugin', formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument('-d',
                            '--database',
                            action='store',
                            help='Database file. Can be the local database name for SQLite, or an SQLAlchemy compatible URL such as postgresql+psycopg2://dirkjan@/roadtools',
                            default='roadrecon.db')
        add_args(parser)
        args = parser.parse_args()
    db_url = database.parse_db_argument(args.database)
    if args.database_password is None:
        args.database_user, args.database_password = BloodHoundPlugin.detect_db_config()
        if args.database_password is None:
            print('Error: Could not autodetect the Neo4j database credentials from your BloodHound config. Please specify them manually')
            return
    session = database.get_session(database.init(dburl=db_url))
    plugin = BloodHoundPlugin(session, args.neodatabase, args.database_user, args.database_password)
    plugin.main()

if __name__ == '__main__':
    main()
